#include "process-status.h"
#include "process-util.h"

static uid_t P_uid;
static int P_pid;
static int P_ppid;
static char P_stat[COMM_LEN];
static char P_cmd_short[COMM_LEN];
static char P_user[COMM_LEN];
static char P_cmd_long[COMM_LEN];

static int read_from_stat(int pid) {
        char buf[PATH_MAX];
        char cmd_path[PATH_MAX];
        char pathname[PATH_MAX];
        int fd = 0;
        struct stat st;
        int r = 0;

        memset(buf, 0, sizeof(buf));
        memset(cmd_path, 0, sizeof(cmd_path));
        memset(pathname, 0, sizeof(pathname));

        r = snprintf(pathname, sizeof(pathname), "/proc/%d", pid);
        if (r <= 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't snprintf /proc/%d.", pid);
                return -1;
        }

        if (stat(pathname, &st) != 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't stat %s.", pathname);
                return -1;
        }

        P_uid = st.st_uid;

        r = snprintf(buf, sizeof(buf), "/proc/%d/stat", pid);
        if (r <= 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't snprintf /proc/%d/stat.", pid);
                return -1;
        }

        fd = open(buf, O_RDONLY, 0);
        if (fd == -1) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't open %s.", buf);
                return -1;
        }

        r = read(fd, buf, sizeof(buf) - 1);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't read /proc/%d/stat.", pid);
                close(fd);
                return -1;
        }

        r = sscanf(buf, "%d %s %s %d", &P_pid, P_cmd_short, P_stat, &P_ppid);
        if (r < 0) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL,
                        "Can't run sscanf.");
                close(fd);
                return -1;
        }

        close(fd);

        if(P_pid != pid)
                return -1;

        r = snprintf(cmd_path, sizeof(cmd_path), "/proc/%d", pid);
        if (r <= 0) {
            manager_status_printf(NULL, STATUS_TYPE_NORMAL, NULL, "Can't snprintf /proc/%d.", pid);
             return -1;
        }

        /* read from /proc/$pid/cmdline */
        read_cmdline(P_cmd_long, sizeof(P_cmd_long), cmd_path, "cmdline", ' ');

        return 0;
}

static void do_user(void) {
        struct passwd *p = NULL;

        p = getpwuid(P_uid);
        if (p) {
                snprintf(P_user, sizeof(P_user), "%s", p->pw_name);
        } else {
                snprintf(P_user, sizeof(P_user), "%u", P_uid);
        }
}

static void print_proc(void) {
        if ((P_ppid != KTHREADD) && (strcmp(P_cmd_short, "(kthreadd)") != 0)) {
                if (strlen(P_cmd_long) != 0) {
                        manager_status_printf(NULL, STATUS_TYPE_NORMAL,"systemd-shutdown",
                                "%-s\t%-d\t%-d\t%-s", P_user, P_pid, P_ppid, P_cmd_long);
                } else {
                        manager_status_printf(NULL, STATUS_TYPE_NORMAL,"systemd-shutdown",
                                "%-s\t%-d\t%-d\t%-s", P_user, P_pid, P_ppid, P_cmd_short);
                }
        }
}

int process_status(void) {
        manager_status_printf(NULL, STATUS_TYPE_NORMAL,"systemd-shutdown",
                "%s", "-----------------------------------------------------------------");
        manager_status_printf(NULL, STATUS_TYPE_NORMAL,"systemd-shutdown",
                "%s", "USER\tPID\tPPID\tCMD");

        struct dirent *ent = NULL;
        DIR *dir = NULL;

        dir = opendir("/proc");
        if (dir == NULL) {
                manager_status_printf(NULL, STATUS_TYPE_NORMAL,"systemd-shutdown",
                        "%s", "can't open /proc");
                return -1;
        }

        while((ent = readdir(dir))){
                if (*ent->d_name < '0' || *ent->d_name > '9')
                        continue;

                if (read_from_stat(atoi(ent->d_name)) != 0)
                        continue;

                do_user();

                print_proc();
        }

        closedir(dir);

        manager_status_printf(NULL, STATUS_TYPE_NORMAL,"systemd-shutdown",
                "%s", "------------------------------------------------------------------");

        return 0;
}
